#open "Variables";;
#open "Minorant";;

(*Algorithmes de dissemination de donnes*)
(*Variables utilises: besoin de Minorant.ml et de Variables.ml*)

(************************* ALGO 0*******************)
(*Message  diffuser:mthode habituelle par FIFO*)
(*Cot linaire en le nombre de messages*)
let determine_message0 v=
  let r=ref v.(0).Premier and indice=ref 0 in
  for i=1 to !nb_messages-1 do
    if v.(i).Premier< !r then (r:=v.(i).Premier;indice:=i)
  done;
  if !r=max_int then -1 else !indice;;

(*Calcul des proportions idales:sert  1 2 3 et 4**)
let Tau=ref (minorant !nb_canaux !popularite !couts);;
let init_Tau()=
  Tau:=minorant !nb_canaux !popularite !couts;;



(*Note:pour les algos qui suivent, qui implmentent les cots, il se peut qu'aucun message ne soit diffus
   l'instant T. L'algo renvoit alors le nombre -1*)  
(************************* ALGO 1*******************)
(*Probabiliste: message diffus proportionnellement  sqrt p_i*)

(*Calcule les probablits de diffusion et les probas cumules de diffusion*)
let proba=ref (make_vect !nb_messages 0.);;
let proba_cum=ref (make_vect !nb_messages 0.);;
let init_proba()=
  proba:=make_vect !nb_messages 0.;
  proba_cum:=make_vect !nb_messages 0.;
  !proba.(0)<-1./.(float_of_int !nb_canaux *. !Tau.(0));
  !proba_cum.(0)<- !proba.(0);
  for i=1 to !nb_messages-1 do
    !proba.(i)<-1./.(float_of_int !nb_canaux *. !Tau.(i));
    !proba_cum.(i)<- !proba_cum.(i-1)+. !proba.(i);
  done;
  !proba;;
init_proba();;

(*Diffuse alatoirement un message suivant les probas plus haut*)
(*Cot linaire en le nombre de messages*)
let determine_message1()=
  let r=random__float 1. in
  try (
    for i=0 to !nb_messages-1 do
      if !proba_cum.(i)>.r then raise (Trouve i)
    done;
    -1)
  with Trouve i->i;;



(************************* ALGO 2*******************)
(*Drandomatisation gloutonne de 1******************)

(*tat est le tableau des sigmai*)
(*Cot linaire en le nombre de messages*)
let determine_message2 etat=
  let min=ref 0. and indice=ref (-1) in
  for i=0 to !nb_messages-1 do
    let r= !couts.(i)-. !popularite.(i)*.etat.(i)*. !Tau.(i) in
    if r<=. !min then (min:=r;indice:=i)
  done;
  !indice;;  



(****************Tri Fusion**********************)
(*Sert pour 3 et 4. ppe x y renvoie le plus petit des lments*)
let rec fusion ppe= fun
  |[] l2 -> l2
  |l1 [] -> l1
  |(p::q as l1) (u::v as l2)-> if ppe p u then p::(fusion ppe q l2) else u::(fusion ppe l1 v);; 
let rec partage = fun
  |[] -> ([],[])
  |[p]-> ([p],[])
  |(p::q::r) -> let (l1,l2)=partage r in (p::l1,q::l2);;
let rec tri_fusion ppe = fun
  |[]->[]
  |[p]->[p]
  |l->let (l1,l2)=partage l in fusion ppe (tri_fusion ppe l1) (tri_fusion ppe l2);;



(************************* ALGO 3*******************)
(*Nombre d'or***************************************)
let T_nbor=ref 1000.;; (*priode des messages*)

let theta=(1.+.sqrt(5.))/.2.;; (*nb d'or*)
let dec x=x-.floor(x);; (*x mod 1*)

(*tri avec tri_fusion*)
let inf1 x y=x.Place<. y.Place;;

(*Construit l'ordonnancement priodique*)
(*cot T log T o T est la priode*)
let construct_period T=
  (*nonT est la frquence de diffusion du message vide*)
  let nonT=
    let r=ref 0. in
    for i=0 to !nb_messages-1 do
      r:=!r+.1./. !Tau.(i)
    done;
    int_of_float(T*.(float_of_int !nb_canaux -. !r)) in
  (*list_mess contient la liste des messages qu'il faudra trier, chaque message de la liste ayant une valeur
    de type freq_cum*nb d'or mod 1*)
  let list_mess=ref [] in
  for j=1 to nonT do
    list_mess:={Message1=(-1);Place=dec(float_of_int  j *. theta)}:: !list_mess
  done;
  let freq_cum=ref nonT in
  for i=0 to !nb_messages-1 do
    (*ni est la frquence de diffusion du message i*)
    let ni=int_of_float(T/. !Tau.(i)) in
    for j=1 to ni do
      list_mess:={Message1=i;Place=dec(float_of_int (!freq_cum+j) *. theta)}:: !list_mess
    done;
    freq_cum:=!freq_cum+ni  
  done;
  tri_fusion inf1 !list_mess;;

print_string "\nCalcul en cours de l'ordonancement priodique...\n";;
let ordonancement=ref (vect_of_list (construct_period 1.));;
let init_ordonancement n=
  ordonancement:=vect_of_list (construct_period n);;

(*Donne le n_ime message  diffuser*)
(*Cot:1*)
let determine_message3 n=
  let m=vect_length !ordonancement in
  !ordonancement.(n mod m).Message1;;



(************************* ALGO 4*******************)
(*******************Utilisation d'arbres************)
let message_arbre_vide()={Message2= -1;Freq=0};;
let inf2=(fun x y->x.Freq<y.Freq);;

(*Parcours donne la feuille indique par les couleurs tout en les schwitchant*)
let rec parcours arb=match arb with
  |Feuille i->i
  |Arbre (g,d)-> 
    if g.Couleur then (g.Couleur<-false;d.Couleur<-true;parcours g.Branche)
    else (d.Couleur<-false;g.Couleur<-true;parcours d.Branche);;

(*Prend en argument une liste de puissances de 2 (dans Freq) trie par ordre croissant et dont la somme est une puissance de
  deux et renvoit deux listes de puissances de 2 dont les sommes sont gales*)
let partage l=
  let sum=list_it (fun x y->x.Freq+y) l 0 in
  constr l [] 0
where rec constr a b s=
  if s=sum/2 then (a,rev b)
  else match a with
    |[]->failwith "Entrez des puissances de deux par ordre croissant dont la somme est une puissance de deux!"
    |p::q -> constr q (p::b) (s+p.Freq);;     

(*Construit l'arbre  partir d'une liste de puissances de 2 ... *)
let rec construct_arb = fun
  |[p]->Feuille p
  | l ->let (l1,l2)=partage l in Arbre ({Couleur=true;Branche=construct_arb l1},
                                        {Couleur=false;Branche=construct_arb l2});;

(*approche x par la puissance de 2 immdiatement suprieure*)
let approx x=
  let r=puissance 2 (int_of_float (log x /. log 2.)) in
  if r>=int_of_float (ceil x) then r else 2*r;;

(*Transforme un tableau en liste de puissances de 2 ... *)
let construct_message t trou=
  let n=vect_length t in
  let min=ref t.(0) in
  for i=1 to n-1 do
    if t.(i)<. !min then min:=t.(i)
  done;
  for i=0 to n-1 do
    t.(i)<-t.(i)/. !min
  done;
  let l=ref [] and puiss_cum=ref 0 in
  for i=0 to n-1 do
    let r=approx t.(i) in
    l:= {Message2=i;Freq=r} :: !l;
    puiss_cum:=!puiss_cum+r
  done;
  (*on rajoute les messages vides, si les cots de diffusion sont non nuls
    Sinon, on voit qu'il vaut mieux rajouter de vrais messages.
    L'algo comblera les trous par le message trou sauf si trou=-2 auquel cas il
    comble les trous en allant du premier message au dernier - cel fonctionne mieux si la distribution est de Zipf*)
  let rest=ref ((approx (float_of_int !puiss_cum)) - !puiss_cum) in
  let compteur=ref 1 and num=ref 0 in
  while !rest>0 do
    if !rest mod 2=1 then (if trou= -2 then (l:={Message2= !num;Freq= !compteur}:: !l;incr num)
                                      else l:={Message2= trou;Freq= !compteur}:: !l);
    compteur:=!compteur*2;rest:=!rest/2;
  done;
  tri_fusion inf2 !l;;   

let arb=ref (Feuille (message_arbre_vide()));;
let init_arbre()=
  let n=vect_length !Tau in
  let freq=make_vect n 0. in
  for i=0 to n-1 do
    freq.(i)<-1./.(float_of_int !nb_canaux *. !Tau.(i));
  done;
  (*Si les couts de diff ne sont pas nuls, on comble les trous par le message vide*) 
  if !couts=make_vect !nb_messages 0. then arb:=construct_arb (construct_message freq (-2))
                                     else arb:=construct_arb (construct_message freq (-1));
 !arb;;

let determine_message4()=(parcours !arb).Message2;;



(************************* ALGO 5*******************)
(****************GLOUTON Dynamique******************)  

let determine_message5 tabl t= (*tabl reprsente le tableau des files d'attentes, cf simul, t reprsente le temps*)
  let max=ref 0. and indice=ref (-1) in
  for i=0 to !nb_messages-1 do
    let r=sqrt(float_of_int(tabl.(i).Nb_clients*(t+1)) -. tabl.(i).Attente_cum) -. !couts.(i) in
    if r>=. !max then (max:=r;indice:=i)
  done;
  !indice;;
  
(********************** ALGO 6 ********************)
(* Algorithme trs mauvais utilis pour le tltexte*)

(*Cet algo consiste  diffuser cycliquement les messages!
  D'o un temps d'attente horrible gal  nombre de messages/2*)
  
let message_actuel=ref (-1);;
let determine_message6()=
  incr message_actuel;
  !message_actuel mod !nb_messages;;