(************************************************Calcul du minorant*****************************)

let Abs x=
  if x<.0. then -.x else x;;

(*Calcul de lambda dans le minorant, trs rapide->ne pas hsiter  prendre une prcision leve*)
(*Newton fonction, drive, approximation alloue, point de dpart*)
let rec Newton f f' epsilon x=
   if Abs(f(x))<=.epsilon then x
   else Newton f f' epsilon (x-.f(x)/.f'(x));;


(*Calcule le minorant en relaxant les contraintes Taui>=1*)   
(*minorant_relaxe nombre de canaux,table des popularits, table des cots*)
let minorant_relaxe w p c=
  let m=vect_length p in
  let sigma=ref 0. in
  for i=0 to m-1 do
    sigma:=!sigma+.sqrt(p.(i)/.(2.*.c.(i)))
  done;
  (*Calcul de lambda, si jamais sigma n'est pas dans la frontire*)
  let lambda=if !sigma<=float_of_int(w) then 0.
             else ((*Comme la fonction objectif est convexe dcroissante, pour que Newton converge, il faut lui entrer
                     une valeur telle que f x soit < 0*)
                   let essai=ref (float_of_int m /. (2.*.float_of_int (w*w))) in
                   let f=(fun x-> let r=ref 0. in for i=0 to m-1 do r:=!r+.sqrt(p.(i)/.(2.*.c.(i)+.x)) done;
                                  float_of_int w -. !r) in
                   while f !essai >. 0. do essai:=!essai/.100. done;
             (*Appel de Newton avec une prcision de 10^-12*)      
             Newton f  (fun x-> let r=ref 0. in for i=0 to m-1 do
                                                    r:=!r+.sqrt(p.(i))*.(2.*.c.(i)+.x)**(-1.5)/.2. 
                                                done;
                                !r)
                         (10.**(-12.)) !essai) in
  (*Calcul des Tau*)
  let Tau=make_vect m 0. in
  for i=0 to m-1 do
    Tau.(i)<-sqrt((2.*.c.(i)+.lambda)/.p.(i))
  done;
  Tau;;
 
  
(*Le type est utilis par l'algo iter_minorant*)
type message={Popu:float;Cout:float};;

(*Lorsque minorant_relax renvoit un Taui<1, on vire le message et un canal et on itre*)
(*iter_minorant nb_canaux, table des messages*)
let iter_minorant w m_vect=
  let m=vect_length m_vect in
  let Vire=make_vect m 0 in
  let Sol=make_vect m 0. in
  let nb_vire=ref 0 in
  let virer i=
    incr nb_vire;
    for j=i to m-1 do
      Vire.(j)<-Vire.(j)+1
    done in  
  let b=ref false in
  while not !b do
    b:=true;
    (*On construit le tableau de message en enlevant ceux qui ont t virs*)
    let P=make_vect (m- !nb_vire) 0. and C=make_vect (m- !nb_vire) 0. in
    for i=0 to m- !nb_vire-1 do
      P.(i)<-m_vect.(i+Vire.(i)).Popu;
      C.(i)<-m_vect.(i+Vire.(i)).Cout;
    done;
    let Tau=minorant_relaxe (w- !nb_vire) P C in
    for i=0 to m- !nb_vire-1 do
      if Tau.(i)<1. then (Sol.(i+Vire.(i))<-1.;virer (i+Vire.(i));b:=false)
    done;
    if !b then (
      (*On entre les derniers Taui* - ceux qui ne sont pas gaux  1*)
      for i=0 to m- !nb_vire-1 do
        Sol.(i+Vire.(i))<-Tau.(i)
      done
    )  
  done;  
  Sol;;


(*Calcule les Taui, avec ou sans cot*)
(*minorant nb_canaux, popu, couts*)
let minorant w p c=
  let m=vect_length p in
  let vide={Popu=0.;Cout=0.} in
  let vect=make_vect m vide in
  for i=0 to m-1 do
    vect.(i)<-{Popu=p.(i);Cout=c.(i)}
  done;
  iter_minorant w vect;;

let minor_sans_cout w p=
  let m=vect_length p in
  minorant w p (make_vect m 0.);;

(*Calcul le minorant, avec ou sans cot de diffusion*)
let calc_min w p c=
  let m=vect_length p in
  let Tau=minorant w p c in
  let min=ref 1. in
  for i=0 to m-1 do
    min:=!min+.p.(i)*.Tau.(i)/.2.+.c.(i)/.Tau.(i)
  done;
  !min;;   

let calc_min_sans_cout w p=
  let m=vect_length p in
  let Tau=minor_sans_cout w p in
  let min=ref 1. in
  for i=0 to m-1 do
    min:=!min+.p.(i)*.Tau.(i)/.2.
  done;
  !min;;
  
(*Calcule le temps d'attente optimal pour un message donn*)
let calc_min_message w p c i=
  let m=vect_length p in
  let Tau=minorant w p c in
  1.+.Tau.(i)/.2.+.c.(i)/.Tau.(i);;
  