#open "Variables";;
#open "Minorant";;
#open "Algorithmes";;
#open "ChangeDyna";;

(*Simulation du log des clients*)
(*Variables utilises: besoin de Minorant.ml et de Variables.ml (et de launch.ml pour la partie dynamique*)

(**************************************************** Simulation ****************************************************)

(*Utilis pour la simu. Les clients arrivent en un temps continu, d'o le type de Arrivee.
  Numro est utilis pour l'algo FIFO*)
type client={Arrivee : float;Numero : int};;

(*Slection d'un message voulu pour un client*)
let rand_mess()=
  let r=random__float 1. in
  try (
    for i=0 to !nb_messages-1 do
      if !popu_cum.(i)>.r then raise (Trouve i)
    done;
    !nb_messages-1)
  with Trouve i->i;;

(*Gestion des files d'attentes*)  
let file_simul_vide()={Nb_clients=0;Attente_cum=0.;Premier=max_int};;
(*Ajoute un client  une file d'attente. Cot:1*)
let add f c=
  f.Nb_clients<-f.Nb_clients+1;
  f.Attente_cum<-f.Attente_cum+.c.Arrivee;
  if c.Numero<f.Premier then f.Premier<-c.Numero;;

(*Calcule le nombre de clients dans le tableau des files d'attentes*)
let nombre_de_clients v=
  let r=ref 0 in
  for i=0 to !nb_messages-1 do
    r:=!r+v.(i).Nb_clients
  done;
  !r;;    

(*Ajoute l'attente cumule,f reprsentant la file d'attente pour le message diffus*)
(*Cot:1*)
let add_attente f t=
  float_of_int(f.Nb_clients*(t+2)) -. f.Attente_cum;; (*t+2 en prenant en compte le temps de tlchargement*)

(*AJoute l'attente cumule des clients n'ayant pas t servis  la fin de la simulation*)
(*tps_attente tableau des files d'attentes, temps; Cot:linaire en le nombre de messages*)
let tps_attente_restant v t=
  let r=ref 0. in
  for i=0 to !nb_messages-1 do
    r:=!r+.add_attente (v.(i)) t
  done;
  !r;;                                     
(*********************** La simulation: *************************************)

let FIFO=0;;
let RAND=1;;
let GLOUTON=2;;
let NBOR=3;;
let ARBRE=4;;
let DYNA=5;;
let TELETEXT=6;;

let imprimer commentaire i=print_string commentaire;print_float (arrondir i);print_newline();;
let imprimer_int commentaire i=print_string commentaire;print_int i;print_newline();;

(*simul algorithme, longueur de la simulation, nombre de clients se loguant par unit de temps.
  La variation du nombre_de_clients ne fait varier le temps d'attente que pour FIFO et GLOUTON_DYNA
   ->utilit de la dissmination de donnes*)
let simul algo long_simu freq_pers=
  let tabl_attente=make_vect !nb_messages (file_simul_vide()) in
  let etat=make_vect !nb_messages 0. in (*etat sert  glouton, reprsente la dure depuis la dernire diffusion*)
    (*Initialisation de la table d'attente, l'initiation plus haut ne crant qu'une seule rfrence*)
    for z=0 to !nb_messages-1 do
      tabl_attente.(z)<-file_simul_vide()
    done;
  let nb_clients=ref 0 and tmps_attente=ref 0. in (*Pour calculer le temps d'attente moyen*)
    for i=0 to long_simu do
      (*log des clients*)
      for j=1 to freq_pers do
        add tabl_attente.(rand_mess()) {Arrivee=float_of_int(i)+.random__float 1.;Numero= !nb_clients};
        incr nb_clients
      done;  
      (*Actualisation de l'tat*)
      for j=0 to !nb_messages-1 do
        etat.(j)<-etat.(j)+.1.
      done;
      (*Diffusion dans les canaux*)
      for j=0 to !nb_canaux-1 do
        let message=(if algo=1 then determine_message1() else
                     if algo=2 then determine_message2 etat else
                     if algo=3 then determine_message3 (i* !nb_canaux+j) else
                     if algo=4 then determine_message4() else
                     if algo=5 then determine_message5 tabl_attente i else
                     if algo=6 then determine_message6 () else
                                    determine_message0 tabl_attente) in
        if message<> -1 then (
          etat.(message)<-0.;
          tmps_attente:=!tmps_attente+.add_attente (tabl_attente.(message)) i +. !couts.(message);
          tabl_attente.(message)<-file_simul_vide();
        )  
      done;          
    done;
    (*Rsultats*)
    (*On a deux mthodes de calcul du temps de service moyen, qui convergent toutes les deux pour une dure de simu
      assez grande, mais la deuxime est beaucoup plus fiable.
    imprimer "Temps d'attente moyen actuel: " (!tmps_attente /. float_of_int (!nb_clients-nombre_de_clients tabl_attente));*)
    (!tmps_attente+. tps_attente_restant tabl_attente long_simu) /. float_of_int !nb_clients;;
   
    
(*simul2 est exactement comme simul, sauf que le nombre de clients se loguant dans une unit de temps n'est
  plus entier mais probabiliste d'esprance freq_pers.
  simul2 est un peu plus lent, c'est pourquoi j'ai conserv simul1 -importance de la rapidit pour des
  simus se droulant sur un long intervalle de temps
  enfin, simul2 permet galement de prsenter les rsultats par intervalles de longueurs de temps, ce
  qui est utile pour tester la variation du temps d'attente lorsque le comportement des clients change*)
let simul2 algo long_simu freq_pers pallier=
  let client_cum=ref 0. in (*utilis pour calculer le nombre de clients se loguant en une unit de temps*)
  let tabl_attente=make_vect !nb_messages (file_simul_vide()) in
  (*Initialisation de la table d'attente, l'initiation plus haut ne crant qu'une seule rfrence*)
  for z=0 to !nb_messages-1 do
    tabl_attente.(z)<-file_simul_vide()
  done;
  let etat=make_vect !nb_messages 0. in (*etat sert  glouton*)
  let nb_clients=ref 0 and tmps_attente=ref 0. in
    for i=1 to long_simu do
      (*log des clients*)
      client_cum:=!client_cum+.random__float (2.*.freq_pers);
      while !client_cum>.1. do
        add tabl_attente.(rand_mess()) {Arrivee=float_of_int(i)+.random__float 1.;Numero= !nb_clients};
        incr nb_clients;
        client_cum:=!client_cum-.1.
      done; 
      (*Actualisation de l'tat*)
      for j=0 to !nb_messages-1 do
        etat.(j)<-etat.(j)+.1.
      done;
      (*Diffusion dans les canaux*)
      for j=0 to !nb_canaux-1 do
        let message=(if algo=1 then determine_message1() else
                     if algo=2 then determine_message2 etat else
                     if algo=3 then determine_message3 (i* !nb_canaux+j) else
                     if algo=4 then determine_message4() else
                     if algo=5 then determine_message5 tabl_attente i else
                     if algo=6 then determine_message6 () else
                                    determine_message0 tabl_attente) in
        if message<> -1 then (
          etat.(message)<-0.;
          tmps_attente:=!tmps_attente+.add_attente (tabl_attente.(message)) i +. !couts.(message);
          tabl_attente.(message)<-file_simul_vide();
        )  
      done;
      (*rsultats intermdiaires:*)
      if i mod pallier=0 then (
        imprimer_int "Pallier :" i;
        imprimer_int "Clients restant  servir:" (nombre_de_clients tabl_attente);
        imprimer "Attente restante moyenne:" ((tps_attente_restant tabl_attente i)/.float_of_int(nombre_de_clients tabl_attente));
        print_newline());           
    done;
    (*Rsultats*)
    (*imprimer "Temps d'attente moyen actuel: " (!tmps_attente /. float_of_int (!nb_clients-nombre_de_clients tabl_attente));*)
    (!tmps_attente+. tps_attente_restant tabl_attente long_simu) /. float_of_int !nb_clients
    (*tmps_attente reprsente l'attente moyenne de service pour un client, le reste reprsente l'attente moyenne des 
      clients n'ayant pas encore ts servis*);;
      
            
(********************** Attente moyenne thorique ************************)
(*Tous les algos de dissmination de donns n'ont pas besoin de connatre quels clients sont logus.
  L'algo ci_dessous calcule le temps d'attente en arrire moyen. On peut montrer que ce temps converge vers le
  temps de service. C'est ce que l'on observe en pratique lorsqu'on compare thorie et simul pour une longueur de
  simulation d'environ 100*nb_messages
  En fait, theorie calcule le cot du chemin du graphe des tats*)

(*thorie algo  utiliser, longueur de la simulation*)
let theorie algo long=
  (*etat.(i) est le temps d'attente en arrire du message i*)
  let etat=make_vect !nb_messages 0. in 
  let cout_tot=ref 0. in
    for i=0 to long do
      (*Actualisation de l'tat*)
      for j=0 to !nb_messages-1 do
        etat.(j)<-etat.(j)+.1.
      done;
      (*Diffusion dans les canaux*)
      for j=0 to !nb_canaux-1 do
        let message=(if algo=2 then determine_message2 etat else
                     if algo=3 then determine_message3 (i* !nb_canaux+j) else
                     if algo=4 then determine_message4 () else
                     if algo=6 then determine_message6 () else
                                    determine_message1()) in
        if message<> -1 then (etat.(message)<-0.;cout_tot:=!cout_tot+. !couts.(message));
      done;  
      cout_tot:=!cout_tot+.1.5; (*+1 pour le temps de tlchargement, +1/2  cause de la continuit du temps*)
      for j=0 to !nb_messages-1 do
        cout_tot:=!cout_tot +. !popularite.(j)*.etat.(j)
      done;
    done;
    !cout_tot/.float_of_int (long+1);;
    
(*comme simul2, theorie2 est exactement comme theorie sauf qu'il donne des rsultats plus complets
  en particulier il donne la liste des messages diffuser par l'algo et le temps d'attente pour chaque message,
  ce qui permet d'tudier le comportement galitariste ou non des diffrents algos...*)
let theorie2 algo long=
  (*etat.(i) est le temps d'attente en arrire du message i*)
  let etat=make_vect !nb_messages 0. in 
  let cout_tot=ref 0. in
  let messages_diffuss=ref [] in (*contient la liste des messages diffuss*)
  let nb_message_vect=make_vect !nb_messages 0 
  and cout_message=make_vect !nb_messages 0. in (*pour avoir le temps d'attente de chaque message*)
    for i=0 to long do
      (*Actualisation de l'tat*)
      for j=0 to !nb_messages-1 do
        etat.(j)<-etat.(j)+.1.
      done;
      (*Diffusion dans les canaux*)
      for j=0 to !nb_canaux-1 do
        let message=(if algo=2 then determine_message2 etat else
                     if algo=3 then determine_message3 (i* !nb_canaux+j) else
                     if algo=4 then determine_message4 () else
                     if algo=6 then determine_message6 () else
                                    determine_message1()) in
        if message<> -1 then (etat.(message)<-0.;cout_tot:=!cout_tot+. !couts.(message);
                              nb_message_vect.(message)<-nb_message_vect.(message)+1;
                              messages_diffuss:=message:: !messages_diffuss);
      done;  
      cout_tot:=!cout_tot+.1.5; (*+1 pour le temps de tlchargement, +1/2  cause de la continuit du temps*)
      for j=0 to !nb_messages-1 do
        cout_tot:=!cout_tot +. !popularite.(j)*.etat.(j);
        cout_message.(j)<-cout_message.(j) +. etat.(j);
      done;
    done;
    for j=0 to !nb_messages-1 do
      cout_message.(j)<-cout_message.(j) /. float_of_int long +. 1.5 (*+1 pour le temps de tlchargement, +1/2  cause de la continuit du temps*)
    done;
    (rev !messages_diffuss, nb_message_vect, cout_message, !cout_tot/.float_of_int (long+1));;
  
    
let simul_all long freq=
  print_string "\n";
  imprimer "Minorant: " (calc_min !nb_canaux !popularite !couts);
  print_string "FIFO:\n";
    imprimer "  Temps d'attente moyen asymptotique: " (simul FIFO long freq);
  print_string "RAND:\n";
    imprimer "  Thorie: " (theorie RAND long);
    imprimer "  Temps d'attente moyen asymptotique: " (simul RAND long freq);
  print_string "GLOUTON:\n";
    imprimer "  Thorie: " (theorie GLOUTON long);
    imprimer "  Temps d'attente moyen asymptotique: " (simul GLOUTON long freq);
  print_string "NB OR:\n";
    imprimer "  Thorie: " (theorie NBOR long);
    imprimer "  Temps d'attente moyen asymptotique: " (simul NBOR long freq);
  print_string "ARBRE:\n";
    imprimer "  Thorie: " (theorie ARBRE long);
    imprimer "  Temps d'attente moyen asymptotique: " (simul ARBRE long freq);
  print_string "GLOUTON DYNAMIQUE:\n";
    imprimer "  Temps d'attente moyen asymptotique: " (simul DYNA long freq);
  print_string "TELETEXTE:\n";
    imprimer "  Thorie: " (theorie TELETEXT long);
    imprimer "  Temps d'attente moyen asymptotique: " (simul TELETEXT long freq);;

(*Calcul exact du temps d'attente d'un ordonnancement priodique sur un canal*)
let calc_ts v i=
  let n=vect_length v in
  let premier=ref (-1) in
  let total=ref 0. and t=ref 0 in
  for j=0 to n-1 do
    if v.(j)=i then (
      if !premier= -1 then (premier:=j;t:=j)
      else (
        let delta=float_of_int(j- !t) in
        t:=j;
        total:=!total +. delta*.delta +. !couts.(i)
      )
    )
  done;
  if !premier= -1 then total:=1./.0. else total:=!total +. float_of_int(!premier+n- !t) ** 2.;
  !total/.float_of_int(2*n) +. 1.;;
 
let calc v=
  let r=ref 0. in
  for i=0 to !nb_messages-1 do
    r:=!r +. !popularite.(i) *. calc_ts v i
  done;
  !r;;

let calc_algo algo periode=
  match (theorie2 algo periode) with (a,b,c,d)->calc (vect_of_list a);;
     
(****************************Simulation dynamique******************************************)   
(*simul_dyna permet de simuler des variations de popularit*) 

let simul_dyna popu_vect algo longueur freq_pers pallier=
  
  let sortie1=open_out ("C:\Temp\ " ^(string_of_int algo)^ "Pallier.txt") in
  let sortie2=open_out ("C:\Temp\ " ^(string_of_int algo)^ "ClientsRestants.txt") in
  let sortie3=open_out ("C:\Temp\ " ^(string_of_int algo)^ "TmpsAttente.txt") in
  let sortie4=open_out ("C:\Temp\ " ^(string_of_int algo)^ "Popularit.txt") in
 

  
  let m=vect_length popu_vect in
  nb_messages:=vect_length popu_vect.(0).Popu;
  let duree=ref 0 in
  for i=0 to m-1 do
    duree:=!duree+popu_vect.(i).Duree
  done;
  (*Calcul de la popularit moyenne*)
  let popu_moyen=make_vect !nb_messages 0. in
  for i=0 to !nb_messages-1 do
    popu_moyen.(i)<-
      (let r=ref 0. in
       for j=0 to m-1 do
         r:=!r+.popu_vect.(j).Popu.(i)*.float_of_int(popu_vect.(j).Duree)/.(float_of_int !duree)
       done;
       !r)
  done;
  change_popu popu_moyen;
  (*tabl_temps_attente et tabl_clients permettent de calculer le temps de service  chaque changement 
    de la distribution des popularits*)
  let tabl_temps_attente=make_vect (m+1) 0. and tabl_clients=make_vect (m+1) 0 in
  
  (*Simulation proprement dite, base sur Simul2
    On ne peut utiliser directement simul2 car cel reviendrait  vider les files d'attentes 
    chaque changement de popularit*)
  let client_cum=ref 0. in (*utilis pour calculer le nombre de clients se loguant en une unit de temps*)
  let tabl_attente=make_vect !nb_messages (file_simul_vide()) in
  (*Initialisation de la table d'attente, l'initiation plus haut ne crant qu'une seule rfrence*)
  for z=0 to !nb_messages-1 do
    tabl_attente.(z)<-file_simul_vide()
  done;
  let etat=make_vect !nb_messages 0. in (*etat sert  glouton*)
  let nb_clients=ref 0 and tmps_attente=ref 0. in
  (*popularit ne sert que pour les algos et le calcul du minorant. Lorsqu'un client se logue, pour savoir quel 
  message il demande, on utilise popu_cum. Donc ici on se sert de popu_moyen pour les algos, et  chaque changement
  de distribution de popularit on change uniquement popu_cum, ce qui change uniquement les habitudes des clients
  dans la simulation*)
  let popu_num=ref 0 in let popu_change=ref (popu_vect.(!popu_num).Duree*longueur/ !duree) in
  for i=1 to longueur do
    (*log des clients*)
    client_cum:=!client_cum+.random__float (2.*.freq_pers);
    while !client_cum>.1. do
      add tabl_attente.(rand_mess()) {Arrivee=float_of_int(i)+.random__float 1.;Numero= !nb_clients};
      incr nb_clients;
      client_cum:=!client_cum-.1.
    done;   
    (*Actualisation de l'tat*)
    for j=0 to !nb_messages-1 do
      etat.(j)<-etat.(j)+.1.
    done;
    (*Diffusion dans les canaux*)
    for j=0 to !nb_canaux-1 do
      let message=(if algo=1 then determine_message1 () else
                   if algo=2 then determine_message2 etat else
                   if algo=3 then determine_message3 (i* !nb_canaux+j) else
                   if algo=4 then determine_message4 () else
                   if algo=5 then determine_message5 tabl_attente i else
                   if algo=6 then determine_message6 () else
                                    determine_message0 tabl_attente) in
      if message<> -1 then (
        etat.(message)<-0.;
        tmps_attente:=!tmps_attente+.add_attente (tabl_attente.(message)) i +. !couts.(message);
        tabl_attente.(message)<-file_simul_vide();
      )  
    done;
    (*rsultats intermdiaires:*)
    if i mod pallier=0 then (
      imprimer_int "Pallier :" i;
      imprimer_int "Clients restant  servir:" (nombre_de_clients tabl_attente);
      imprimer "Attente restante moyenne:" ((tps_attente_restant tabl_attente i)/.float_of_int(nombre_de_clients tabl_attente));
      print_newline();
      
    output_string sortie1 (string_of_int i);output_string sortie1 "\n";
    output_string sortie2 (string_of_int ((nombre_de_clients tabl_attente)));output_string sortie2 "\n";
    output_string sortie3 (string_of_float ((tps_attente_restant tabl_attente i)/.float_of_int(nombre_de_clients tabl_attente)));output_string sortie3 "\n"
    );
    
    (*changement de la popularit des messages:*)
    if i= !popu_change then (
      tabl_clients.(!popu_num+1)<- !nb_clients - tabl_clients.(!popu_num)-(nombre_de_clients tabl_attente);
      tabl_temps_attente.(!popu_num+1)<- !tmps_attente+. (tps_attente_restant tabl_attente i)-. tabl_temps_attente.(!popu_num);
      imprimer ("  Popularit n"^string_of_int (!popu_num+1)^" ") 
               (tabl_temps_attente.(!popu_num+1)/. (float_of_int (tabl_clients.(!popu_num+1)+(nombre_de_clients tabl_attente))));   
      print_newline();
      
      output_string sortie4 (string_of_float (tabl_temps_attente.(!popu_num+1)/. (float_of_int (tabl_clients.(!popu_num+1)+(nombre_de_clients tabl_attente)))));output_string sortie4 "\n";
      
      if i<>longueur then (
        popu_cum:=cumul(normalise popu_vect.(!popu_num).Popu);
        incr popu_num;popu_change:= !popu_change + popu_vect.(!popu_num).Duree*longueur/ !duree))          
  done;
  (*Rsultats*)  
  output_string sortie4 (string_of_float ((!tmps_attente+. (tps_attente_restant tabl_attente longueur)) /. (float_of_int !nb_clients)));output_string sortie4 "\n";
  close_out sortie1;
  close_out sortie2;
  close_out sortie3;
  close_out sortie4;

  (!tmps_attente+. (tps_attente_restant tabl_attente longueur)) /. (float_of_int !nb_clients);;
  
let simul_all_dyna popu_vect longueur nb_clients pallier=
  print_string "FIFO:\n";
    imprimer "  Temps d'attente moyen asymptotique: " (simul_dyna popu_vect FIFO longueur nb_clients pallier);
  print_string "\nRAND:\n";
    imprimer "  Temps d'attente moyen asymptotique: " (simul_dyna popu_vect RAND longueur nb_clients pallier);
  print_string "\nGLOUTON:\n";
    imprimer "  Temps d'attente moyen asymptotique: " (simul_dyna popu_vect GLOUTON longueur nb_clients pallier);
  print_string "\nNB OR:\n";
    imprimer "  Temps d'attente moyen asymptotique: " (simul_dyna popu_vect NBOR longueur nb_clients pallier);
  print_string "\nARBRE:\n";
    imprimer "  Temps d'attente moyen asymptotique: " (simul_dyna popu_vect ARBRE longueur nb_clients pallier);
  print_string "\nGLOUTON DYNAMIQUE:\n";
    imprimer "  Temps d'attente moyen asymptotique: " (simul_dyna popu_vect DYNA longueur nb_clients pallier);;     