(**************Variable contient les types, algo ou constantes utilises par plusieurs des modules*********)

(*Serveur*)
let nb_canaux= ref 1;;
let nb_messages= ref 1;;


(*Calcul des cots et des popularits, dans un tableau*)
(*Messages de mme longueur, distribution suivant la loi de Zipf*)
let popularite=ref (make_vect !nb_messages 0.);; (*la longueur de popularite peut changer*) 
let popu_cum=ref (make_vect !nb_messages 0.);;   (*ce qui explique pourquoi on prend une rfrence*)
let duree=ref (make_vect !nb_messages 1.);;
let couts=ref (make_vect !nb_messages 0.);;


(*Construction des tableaux*)
(*Construction d'une tableau de somme cumule  partir d'un tableau de float*)
let cumul p=
  let n=vect_length p in
  let p_cum=make_vect n 0. in
  p_cum.(0)<- p.(0);
  for i=1 to n-1 do
    p_cum.(i)<- p_cum.(i-1)+. p.(i);
  done;
  p_cum;;  

(*Normalisation*)
let normalise p=
  let n=vect_length p in
  let r=ref 0. in
  for i=0 to n-1 do
    r:=!r+. p.(i)
  done;
  for i=0 to n-1 do
    p.(i)<-p.(i)/. !r
  done;
  p;;

(*Cre un tableau suivant la loi de Zipf:le nieme message a une popularite proportionelle  1/n*)
let zipf n=
  let l=make_vect n 0. in
  for i=0 to n-1 do
    l.(i)<-1./.(float_of_int(i+1))
  done;
  l;; 

(*unif cre un tableau de n lments qui diffrent au plus de epsilon,sert pour avoir des popularits uniformes*)
let unif n epsilon=
  let l=make_vect n 0. in 
  let r=ref 0. in
  for i=0 to n-1 do
    l.(i)<-1./.(float_of_int n)+.random__float epsilon -. epsilon;
    if l.(i)<.0. then l.(i)<-0.;
    r:=!r+.l.(i)
  done;
  l;;
  
(*Cre un ensemble avec des popularits leves, et l'autre avec des popus faibles*)
let disparate n l a=
  let v=make_vect n 0. in
  for i=0 to (min n l)-1 do 
    v.(i)<-a
  done;
  for i=l to n-1 do
    v.(i)<-1.
  done;
  v;; 
  
let init_popularite()=
  popularite:=normalise(zipf !nb_messages);
  popu_cum:=cumul !popularite;;
init_popularite();;

let init_duree()=
  duree:=make_vect !nb_messages 1.;;

let init_couts()=
  couts:=make_vect !nb_messages 0.;;

(*Sous-prog utiliss*)
let arrondir x=floor(100.*.x)/.100.;;
let rec puissance x y=
  let r=if y>0 then puissance x (y/2) else 1 in
  if y mod 2 = 1 then x*r*r
  else r*r;;